package org.openmuc.jdlms.internal.lnassociation;

import java.io.IOException;
import java.util.Arrays;
import java.util.Map;

import org.openmuc.jdlms.Authentication;
import org.openmuc.jdlms.LogicalDevice;
import org.openmuc.jdlms.internal.APdu;
import org.openmuc.jdlms.internal.AssociateSourceDiagnostic.AcseServiceUser;
import org.openmuc.jdlms.internal.AuthenticationMechanism;
import org.openmuc.jdlms.internal.ConnectionData;
import org.openmuc.jdlms.internal.ContextId;
import org.openmuc.jdlms.internal.DataDirectory.DlmsLogicalDevice;
import org.openmuc.jdlms.internal.ObjectIdentifier;
import org.openmuc.jdlms.internal.asn1.cosem.COSEMpdu;
import org.openmuc.jdlms.internal.asn1.cosem.Conformance;
import org.openmuc.jdlms.internal.asn1.cosem.InitiateRequest;
import org.openmuc.jdlms.internal.asn1.iso.acse.AARQ_apdu;
import org.openmuc.jdlms.internal.asn1.iso.acse.ACSE_apdu;
import org.openmuc.jdlms.internal.asn1.iso.acse.Mechanism_name;
import org.openmuc.jdlms.internal.security.HlsProcessorGmac;
import org.openmuc.jdlms.internal.security.HlsSecretProcessor;
import org.openmuc.jdlms.internal.security.RandomSequenceGenerator;

public class InitialmessageProcessor {

    private final ConnectionData connectionData;
    private final DlmsLogicalDevice dlmsLogicalDevice;
    private final LogicalDevice logicalDevice;

    public InitialmessageProcessor(ConnectionData connectionData, DlmsLogicalDevice dlmsLogicalDevice) {
        this.connectionData = connectionData;
        this.dlmsLogicalDevice = dlmsLogicalDevice;
        this.logicalDevice = dlmsLogicalDevice.getLogicalDevice();
    }

    public APdu processInitialMessage(byte[] messageData) throws IOException, GenericAssociationException {
        InitialResponseBuilder initialResponseBuilder = new InitialResponseBuilder(conformance());

        if (dlmsLogicalDevice == null) {
            throw new AssociatRequestException(AcseServiceUser.NO_REASON_GIVEN);
        }

        LogicalDevice logicalDevice = dlmsLogicalDevice.getLogicalDevice();
        Map<Integer, Authentication> restrictions = logicalDevice.getRestrictions();
        if (restrictions.isEmpty()) {
            this.connectionData.authenticated = true;
            return initialResponseBuilder.setContextId(ContextId.LOGICAL_NAME_REFERENCING_NO_CIPHERING).build();
        }

        Authentication authentication = this.connectionData.authentication;
        if (authentication == null) {
            // unknown client ID
            throw new AssociatRequestException(AcseServiceUser.NO_REASON_GIVEN);
        }

        APdu aPdu;

        if (authentication.getCryptographicAlgorithm() != null) {
            if (this.connectionData.clientSystemTitle == null) {
                this.connectionData.clientSystemTitle = systemTitle();
            }
            aPdu = APdu.decode(messageData, this.connectionData.clientSystemTitle, connectionData.frameCounter,
                    authentication.getGlobalEncryptionKey(), authentication.getAuthenticationKey());
        }
        else {
            aPdu = APdu.decode(messageData);
        }
        if (aPdu.cosemPdu == null) {
            throw new AssociatRequestException(AcseServiceUser.NO_REASON_GIVEN);
        }

        COSEMpdu cosemPdu = aPdu.cosemPdu;

        if (cosemPdu.getChoiceIndex() != COSEMpdu.Choices.INITIATEREQUEST) {
            throw new AssociatRequestException(AcseServiceUser.NO_REASON_GIVEN);
        }
        InitiateRequest initiateRequest = cosemPdu.initiateRequest;
        this.connectionData.clientMaxReceivePduSize = initiateRequest.client_max_receive_pdu_size.getValue() & 0xFFFF;

        ACSE_apdu acseAPdu = aPdu.acseAPdu;

        if (acseAPdu == null) {
            throw new AssociatRequestException(AcseServiceUser.NO_REASON_GIVEN);
        }
        AARQ_apdu aarq = acseAPdu.aarq;

        return tryToAuthenticate(initialResponseBuilder, aarq, authentication);
    }

    private void checkContextId(ContextId contextId) throws AssociatRequestException {
        boolean snConnection = contextId == ContextId.SHORT_NAME_REFERENCING_NO_CIPHERING
                || contextId == ContextId.SHORT_NAME_REFERENCING_WITH_CIPHERING;
        if (snConnection) {
            throw new AssociatRequestException(AcseServiceUser.APPLICATION_CONTEXT_NAME_NOT_SUPPORTED);
        }
    }

    private APdu tryToAuthenticate(InitialResponseBuilder initialResponseBuilder, AARQ_apdu aarq,
            Authentication authentication) throws AssociatRequestException, IOException {
        ContextId contextId = ObjectIdentifier.applicationContextIdFor(aarq.application_context_name);

        checkContextId(contextId);

        Mechanism_name mechanismName = aarq.mechanism_name;

        if (mechanismName == null && authentication.getAuthenticationMechanism() != AuthenticationMechanism.NONE) {
            throw new AssociatRequestException(AcseServiceUser.AUTHENTICATION_MECHANISM_NAME_REQUIRED);
        }
        else if (mechanismName == null && authentication.getAuthenticationMechanism() == AuthenticationMechanism.NONE) {
            this.connectionData.authenticated = true;
            this.connectionData.mechanismId = AuthenticationMechanism.NONE;
            return initialResponseBuilder.build();
        }

        if (mechanismName == null) {
            throw new AssociatRequestException(AcseServiceUser.NO_REASON_GIVEN);
        }

        this.connectionData.clientToServerChallenge = aarq.calling_authentication_value.charstring.value;

        if (contextId == ContextId.LOGICAL_NAME_REFERENCING_WITH_CIPHERING) {
            // TODO
        }

        AuthenticationMechanism authenticationMechanismId = ObjectIdentifier.mechanismIdFor(mechanismName);
        this.connectionData.mechanismId = authenticationMechanismId;

        if (authenticationMechanismId == AuthenticationMechanism.NONE
                && authentication.getAuthenticationMechanism() != AuthenticationMechanism.NONE) {
            this.connectionData.authenticated = true;
            return initialResponseBuilder.setContextId(ContextId.LOGICAL_NAME_REFERENCING_NO_CIPHERING).build();
        }

        if (authenticationMechanismId != authentication.getAuthenticationMechanism()) {
            throw new AssociatRequestException(AcseServiceUser.AUTHENTICATION_FAILURE);
        }

        switch (authenticationMechanismId) {
        case LOW:
            return processLowAuthentciationRequest(aarq, authentication.getAuthenticationKey());

        case HLS5_GMAC:
            return processHls5GmacAuthentciationRequest(aarq, authentication);
        default:
            throw new AssociatRequestException(AcseServiceUser.APPLICATION_CONTEXT_NAME_NOT_SUPPORTED);
        }
    }

    private APdu processHls5GmacAuthentciationRequest(AARQ_apdu aarq, Authentication authentication)
            throws IOException, AssociatRequestException {
        byte[] clientToServerChallenge = this.connectionData.clientToServerChallenge;

        this.connectionData.clientSystemTitle = aarq.calling_AP_title.ap_title_form2.value;
        byte[] clientSystemTitle = this.connectionData.clientSystemTitle;

        int challengeLength = clientToServerChallenge.length;

        checkChallangeLength(challengeLength);

        this.connectionData.frameCounter = 1;

        byte[] serverToClientChallenge = RandomSequenceGenerator.generate(challengeLength);

        HlsSecretProcessor hlSecretProcessor = new HlsProcessorGmac();

        this.connectionData.processedServerToClientChallenge = hlSecretProcessor.process(serverToClientChallenge,
                authentication.getAuthenticationKey(), authentication.getGlobalEncryptionKey(), clientSystemTitle,
                ++this.connectionData.frameCounter);

        return new InitialResponseBuilder(conformance()).setContextId(ContextId.LOGICAL_NAME_REFERENCING_NO_CIPHERING)
                .setAuthenticationValue(serverToClientChallenge)
                .setSystemTitle(systemTitle())
                .build();

    }

    private void checkChallangeLength(int challengeLength) throws AssociatRequestException {
        if (challengeLength < 8 || challengeLength > 64) {
            throw new AssociatRequestException(AcseServiceUser.AUTHENTICATION_FAILURE);
        }
    }

    private APdu processLowAuthentciationRequest(AARQ_apdu aarq, byte[] authenticationKey)
            throws AssociatRequestException {
        byte[] clientAuthenticaionValue = aarq.calling_authentication_value.charstring.value;

        if (Arrays.equals(clientAuthenticaionValue, authenticationKey)) {
            this.connectionData.authenticated = true;
            return new InitialResponseBuilder(conformance()).build();
        }

        throw new AssociatRequestException(AcseServiceUser.AUTHENTICATION_FAILURE);
    }

    private Conformance conformance() {
        return this.logicalDevice.getConformance();
    }

    private byte[] systemTitle() {
        return this.logicalDevice.getSystemTitle();
    }

}
