/*
 * Copyright 2012-16 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms.app;

import java.io.IOException;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.List;

import org.openmuc.jdlms.Authentication;
import org.openmuc.jdlms.Authentication.CryptographicAlgorithm;
import org.openmuc.jdlms.ConnectionBuilder;
import org.openmuc.jdlms.HexConverter;
import org.openmuc.jdlms.SerialConnectionBuilder;
import org.openmuc.jdlms.TcpConnectionBuilder;
import org.openmuc.jdlms.internal.AuthenticationMechanism;
import org.openmuc.jdlms.internal.cli.CliParameter;
import org.openmuc.jdlms.internal.cli.CliParameterBuilder;
import org.openmuc.jdlms.internal.cli.CliParseException;
import org.openmuc.jdlms.internal.cli.CliParser;
import org.openmuc.jdlms.internal.cli.FlagCliParameter;
import org.openmuc.jdlms.internal.cli.IntCliParameter;
import org.openmuc.jdlms.internal.cli.LongCliParameter;
import org.openmuc.jdlms.internal.cli.StringCliParameter;

class ConsoleClientCliParser {

    CliParser cliParser;

    // common options

    private final IntCliParameter logicalDeviceAddress = new CliParameterBuilder("-ld")
            .setDescription(
                    "The address of the logical device inside the server to connect to. This address is also referred to as the server wPort.")
            .buildIntegerParameter("logical_device_address", 1);
    private final IntCliParameter clientId = new CliParameterBuilder("-cid")
            .setDescription(
                    "The client ID which identifies the client. This ID is also referred to as the client access point or client wPort. 16 = public user.")
            .buildIntegerParameter("client_ID", 16);
    private final FlagCliParameter useShortNameReferencing = new CliParameterBuilder("-sn")
            .setDescription(" Use short name referencing instead of long name referencing.").buildFlagParameter();
    private final IntCliParameter securityLevel = new CliParameterBuilder("-sec")
            .setDescription("Security level: 0 = LOWEST; 1 = LOW; 3 = HIGH_MD5; 5 = HIGH_GMAC.")
            .buildIntegerParameter("security_level", 0);
    private final StringCliParameter encryptionKey = new CliParameterBuilder("-ekey")
            .setDescription("The encryption key. Note that this key is sometimes needed for authentication as well.")
            .buildStringParameter("encryption_key");
    private final StringCliParameter authenticationKey = new CliParameterBuilder("-akey")
            .setDescription(
                    "The authentication key. Either a plain text password when used with security level 2 or in hexadecimal form (e.g. e3e3e3).")
            .buildStringParameter("authentication_key");
    private final FlagCliParameter enableEncryption = new CliParameterBuilder("-enc")
            .setDescription("Enable encryption.").buildFlagParameter();
    private final IntCliParameter challengeLength = new CliParameterBuilder("-cl")
            .setDescription(
                    "Set the length of the authentication challenge, from 8 to 64 byte. For example: kaifa meters maximal 16 bytes.")
            .buildIntegerParameter("challenge_length", 16);
    private final IntCliParameter responseTimeout = new CliParameterBuilder("-rt")
            .setDescription("Max time the clients waits for an answer from the remote meter.")
            .buildIntegerParameter("response_timeout", 20_000);
    private final StringCliParameter manufacturerId = new CliParameterBuilder("-mid").setDescription("Manufacturer id.")
            .buildStringParameter("manufacturer_id", "MMM");
    private final LongCliParameter deviceId = new CliParameterBuilder("-did").setDescription("Unique device id.")
            .buildLongParameter("device_id", 1);

    // TCP options:

    private final StringCliParameter host = new CliParameterBuilder("-h")
            .setDescription("The address of the device you want to access.")
            .setMandatory()
            .buildStringParameter("host");

    private final IntCliParameter port = new CliParameterBuilder("-p").setDescription("The port to connect to.")
            .buildIntegerParameter("port", 4059);

    private final FlagCliParameter useHdlc = new CliParameterBuilder("-hdlc")
            .setDescription("Use HDCL layer instead of wrapper layer.").buildFlagParameter();

    // Serial options:

    private final StringCliParameter serialPort = new CliParameterBuilder("-sp")
            .setDescription(
                    "The serial port used for communication. Examples are /dev/ttyS0 (Linux) or COM1 (Windows).")
            .setMandatory()
            .buildStringParameter("serial_port");

    private final IntCliParameter baudRate = new CliParameterBuilder("-bd")
            .setDescription("Baud rate of the serial port.").buildIntegerParameter("baud_rate", 9600);

    private final LongCliParameter baudRateChangeDelay = new CliParameterBuilder("-d")
            .setDescription(
                    "Delay of baud rate change in ms. USB to serial converters often require a delay of up to 250ms.")
            .buildLongParameter("baud_rate_change_delay", 0);

    private final FlagCliParameter enableBaudRateHandShake = new CliParameterBuilder("-eh")
            .setDescription("Enables the baudrate handshake process.").buildFlagParameter();

    private final StringCliParameter iec21Address = new CliParameterBuilder("-iec")
            .setDescription("Device address, optional field, manufacturer-specific, 32 characters maximum.")
            .buildStringParameter("iec_21_address");

    private final IntCliParameter physicalDeviceAddress = new CliParameterBuilder("-pd")
            .setDescription("The HDLC physical device address.")
            .setMandatory()
            .buildIntegerParameter("physical_dev_address", 0);

    ConsoleClientCliParser() {
        List<CliParameter> commonParameters = new ArrayList<CliParameter>();
        commonParameters.add(logicalDeviceAddress);
        commonParameters.add(clientId);
        commonParameters.add(useShortNameReferencing);
        commonParameters.add(securityLevel);
        commonParameters.add(encryptionKey);
        commonParameters.add(authenticationKey);
        commonParameters.add(enableEncryption);
        commonParameters.add(challengeLength);
        commonParameters.add(responseTimeout);
        commonParameters.add(manufacturerId);
        commonParameters.add(deviceId);

        List<CliParameter> tcpParameters = new ArrayList<CliParameter>();
        tcpParameters.add(host);
        tcpParameters.add(port);
        tcpParameters.addAll(commonParameters);
        tcpParameters.add(useHdlc);
        // TODO
        // tcpParameters.add(physicalDeviceAddress);

        List<CliParameter> serialParameters = new ArrayList<CliParameter>();
        serialParameters.add(serialPort);
        serialParameters.addAll(commonParameters);
        serialParameters.add(physicalDeviceAddress);
        serialParameters.add(baudRate);
        serialParameters.add(baudRateChangeDelay);
        serialParameters.add(enableBaudRateHandShake);
        serialParameters.add(iec21Address);

        cliParser = new CliParser("jdlms-console-client", "DLMS/COSEM client application to access meters");
        cliParser.addParameterGroup("tcp", tcpParameters);
        cliParser.addParameterGroup("serial", serialParameters);
    }

    public void parse(String[] args) throws CliParseException {
        cliParser.parseArguments(args);
    }

    public ConsoleApp connectAndCreateConsoleApp() throws CliParseException, IOException {

        ConsoleApp app;

        switch (cliParser.getSelectedGroup()) {
        case "tcp":
            app = tcpConnect();
            break;
        case "serial":
            app = serialConnect();
            break;
        default:
            throw new IllegalArgumentException("Unknown connection type");
        }

        System.out.printf("** Successfully connected to host: %n");// TODO

        return app;
    }

    private ConsoleApp tcpConnect() throws CliParseException, IOException {

        TcpConnectionBuilder connectionBuilder = new TcpConnectionBuilder(InetAddress.getByName(host.getValue()))
                .setTcpPort(port.getValue());

        if (useHdlc.isSelected()) {
            connectionBuilder.useHldc();
        }

        return setCommonParameters(connectionBuilder);

    }

    private ConsoleApp serialConnect() throws CliParseException, IOException {
        SerialConnectionBuilder connectionBuilder = new SerialConnectionBuilder(serialPort.getValue())
                .setBaudRate(baudRate.getValue())
                .setBaudRateChangeTime(baudRateChangeDelay.getValue())
                .setPhysicalDeviceAddress(physicalDeviceAddress.getValue())
                .disableHandshake();

        if (iec21Address.isSelected()) {
            connectionBuilder.setIec21Address(iec21Address.getValue());
        }

        if (enableBaudRateHandShake.isSelected()) {
            connectionBuilder.enableHandshake();
        }

        return setCommonParameters(connectionBuilder);
    }

    private ConsoleApp setCommonParameters(ConnectionBuilder<?> connectionBuilder)
            throws CliParseException, IOException {
        connectionBuilder.setLogicalDeviceId(logicalDeviceAddress.getValue())
                .setResponseTimeout(responseTimeout.getValue())
                .setClientId(clientId.getValue())
                .setChallengeLength(challengeLength.getValue())
                .setSystemTitle(manufacturerId.getValue(), deviceId.getValue());

        setSecurityLevel(connectionBuilder);

        if (useShortNameReferencing.isSelected()) {
            return new SnConsoleApp(connectionBuilder.buildSnConnection());
        }
        else {
            return new LnConsoleApp(connectionBuilder.buildLnConnection());
        }
    }

    private void setSecurityLevel(ConnectionBuilder<?> connectionBuilder) throws CliParseException {

        byte[] encryptionkey = null;
        CryptographicAlgorithm cryptographicAlgorithm = null;
        if (enableEncryption.isSelected()) {

            encryptionkey = HexConverter.fromShortHexString(encryptionKey.getValue());
            cryptographicAlgorithm = CryptographicAlgorithm.AES_GMC_128;
        }
        Authentication authentication;

        switch (AuthenticationMechanism.getInstance(securityLevel.getValue())) {
        case NONE:
            authentication = Authentication.newNoneAuthentication();
            break;
        case LOW:
            authentication = Authentication.newPasswordAuthentication(readAuthenticationKey(), encryptionkey,
                    cryptographicAlgorithm);
            break;
        case HLS3_MD5:
            authentication = Authentication.newMd5Authentication(readAuthenticationKey(), encryptionkey,
                    cryptographicAlgorithm);
            break;
        case HLS5_GMAC:
            authentication = Authentication.newGmacAuthentication(readAuthenticationKey(), encryptionkey,
                    cryptographicAlgorithm);
            break;
        default:
            throw new CliParseException("Unknown security level: " + encryptionKey.getValue());
        }

        connectionBuilder.setAuthentication(authentication);
    }

    private byte[] readAuthenticationKey() {
        return HexConverter.fromShortHexString(authenticationKey.getValue());
    }

    public void printUsage() {
        System.out.println(cliParser.getUsageString());
    }

}
