package org.openmuc.jdlms.internal.sessionlayer.hdlc;

import static org.openmuc.jdlms.JDlmsException.ExceptionId.HDLC_CONNECTION_CLOSE_ERROR;
import static org.openmuc.jdlms.JDlmsException.ExceptionId.HDLC_CONNECTION_ESTABLISH_ERROR;
import static org.openmuc.jdlms.JDlmsException.Fault.SYSTEM;

import java.io.IOException;
import java.net.InetAddress;
import java.net.SocketException;
import java.text.MessageFormat;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.TimeUnit;

import org.openmuc.jdlms.FatalJDlmsException;
import org.openmuc.jdlms.JDlmsException.ExceptionId;
import org.openmuc.jdlms.JDlmsException.Fault;
import org.openmuc.jdlms.internal.ConfirmedMode;
import org.openmuc.jdlms.internal.connectionsettings.HdlcSettings;
import org.openmuc.jdlms.internal.connectionsettings.HdlcTcpSettings;
import org.openmuc.jdlms.internal.connectionsettings.SerialSettings;
import org.openmuc.jdlms.internal.connectionsettings.Settings;
import org.openmuc.jdlms.internal.transportlayer.LocalDataExchangeConnection;
import org.openmuc.jdlms.internal.transportlayer.TcpLayer;
import org.openmuc.jdlms.internal.transportlayer.TransportLayer;

import gnu.io.CommPortException;

class HdlcDispatcher {
    private static HdlcDispatcher instance;

    private final Map<Object, HdlcConnection> hdlcConnectionMap;

    private HdlcDispatcher() {
        this.hdlcConnectionMap = new HashMap<>();
    }

    public static synchronized HdlcDispatcher instance() {
        if (instance == null) {
            instance = new HdlcDispatcher();
        }
        return instance;
    }

    private class TcpIpEntry {

        private final int port;
        private final InetAddress inetAddress;

        private TcpIpEntry(InetAddress inetAddress, int port) {
            this.inetAddress = inetAddress;
            this.port = port;
        }

        @Override
        public int hashCode() {
            return this.inetAddress.hashCode() ^ this.port;
        }

        @Override
        public boolean equals(Object obj) {
            if (!(obj instanceof TcpIpEntry)) {
                return false;
            }

            TcpIpEntry other = (TcpIpEntry) obj;
            return port == other.port && inetAddress.equals(other.inetAddress);
        }
    }

    public synchronized HdlcConnection connect(HdlcSettings settings, HdlcConnectionListener listener) {

        HdlcConnection hdlcConnection;

        if (settings instanceof HdlcTcpSettings) {
            HdlcTcpSettings tcpSettings = (HdlcTcpSettings) settings;

            hdlcConnection = hdlcConnectionMap.get(new TcpIpEntry(tcpSettings.inetAddress(), tcpSettings.tcpPort()));

            if (hdlcConnection == null) {
                TransportLayer transportLayer = new TcpLayer(tcpSettings);

                hdlcConnection = new HdlcConnection(settings, transportLayer);
            }
        }
        else if (settings instanceof SerialSettings) {
            SerialSettings serialSettings = (SerialSettings) settings;

            hdlcConnection = hdlcConnectionMap.get(serialSettings.serialPortName());

            if (hdlcConnection == null) {
                TransportLayer transportLayer = new LocalDataExchangeConnection(serialSettings);
                hdlcConnection = new HdlcConnection(settings, transportLayer);
            }

        }
        else {
            // TODO: handle this properly.
            throw new UnsupportedOperationException();
        }

        hdlcConnection.registerNewListener(settings.addressPair(), listener);
        return hdlcConnection;
    }

    public class HdlcConnection {
        private final TransportLayer transportLayer;
        private final Settings settings;
        private final Map<HdlcAddressPair, HdlcConnectionListener> listeners;

        private HdlcAddressPair connectionKey;
        private final BlockingQueue<HdlcFrame> incommingQueue;

        private HdlcConnection(Settings settings, TransportLayer transportLayer) {
            this.settings = settings;
            this.transportLayer = transportLayer;
            this.listeners = new LinkedHashMap<>();

            this.incommingQueue = new ArrayBlockingQueue<>(1);
        }

        public synchronized void send(byte[] data) throws IOException {
            this.transportLayer.outpuStream().write(data);
            this.transportLayer.outpuStream().flush();
        }

        public synchronized HdlcParameterNegotiation open(HdlcSettings settings) throws IOException {
            if (this.transportLayer.isClosed()) {
                this.transportLayer.open();

                Thread connectionReader = new Thread(new ConnectionReader());
                connectionReader.setName("HDLC CONNECTION READER");
                connectionReader.start();
            }

            this.connectionKey = settings.addressPair();
            try {
                return connectSequence(settings);
            } catch (IOException ex) {
                removeListenerAndTryClosePhysicalLayer(settings);

                throw ex;
            } finally {
                this.connectionKey = null;
            }
        }

        public synchronized void close(HdlcSettings settings) throws IOException {
            this.connectionKey = settings.addressPair();

            try {
                disconnectSequence(settings);
            } finally {
                removeListenerAndTryClosePhysicalLayer(settings);

                this.connectionKey = null;
            }
        }

        private void removeListenerAndTryClosePhysicalLayer(HdlcSettings settings) throws IOException {
            synchronized (this.listeners) {
                this.listeners.remove(settings.addressPair());

                if (this.listeners.isEmpty()) {
                    this.transportLayer.close();
                }
            }
        }

        private void disconnectSequence(HdlcSettings settings) throws IOException {
            HdlcFrame sendFrame = HdlcFrame.newDisconnectFrame(settings.addressPair(), true);

            try {
                send(sendFrame.encode());
            } catch (FrameInvalidException e) {
                throw new FatalJDlmsException(HDLC_CONNECTION_CLOSE_ERROR, SYSTEM,
                        "Failed to decode disconnect frame.");
            }

            HdlcFrame receivedFrame = waitForFrame(settings.responseTimeout());

            if (receivedFrame == null) {
                throw new FatalJDlmsException(HDLC_CONNECTION_CLOSE_ERROR, SYSTEM,
                        "Didn't receive answer on connection close.");
            }

            if (receivedFrame.frameType() == FrameType.UNNUMBERED_ACKNOWLEDGE
                    || receivedFrame.frameType() == FrameType.DISCONNECT_MODE) {
                // TODO: do something with this information
            }
        }

        private HdlcFrame waitForFrame(long responseTimeout) {
            HdlcFrame receivedFrame = null;
            try {
                receivedFrame = this.incommingQueue.poll(responseTimeout, TimeUnit.MILLISECONDS);
            } catch (InterruptedException e) {
                // ignore, since this should't occur
            }
            return receivedFrame;
        }

        private HdlcParameterNegotiation connectSequence(HdlcSettings settings) throws IOException {

            if (settings.confirmedMode() == ConfirmedMode.UNCONFIRMED) {
                return new HdlcParameterNegotiation(1024, 1);
            }

            HdlcParameterNegotiation dNegotiation = new HdlcParameterNegotiation(
                    HdlcParameterNegotiation.MIN_INFORMATION_LENGTH, HdlcParameterNegotiation.MIN_WINDOW_SIZE);

            HdlcFrame frame = HdlcFrame.newSetNormalResponseModeFrame(settings.addressPair(), dNegotiation, true);

            try {
                send(frame.encode());
            } catch (FrameInvalidException e) {
                throw new FatalJDlmsException(HDLC_CONNECTION_ESTABLISH_ERROR, SYSTEM,
                        "HLDC Frame is invalid. Evaluate cause for details.", e);
            }

            HdlcFrame receiveFrame = waitForFrame(settings.responseTimeout());

            if (receiveFrame == null) {
                throw new FatalJDlmsException(HDLC_CONNECTION_ESTABLISH_ERROR, SYSTEM,
                        "Didn't receieve answer in connecton establish process.");
            }

            switch (receiveFrame.frameType()) {
            case UNNUMBERED_ACKNOWLEDGE:
                if (receiveFrame.informationField() == null) {
                    throw new FatalJDlmsException(HDLC_CONNECTION_ESTABLISH_ERROR, SYSTEM,
                            "Remote meter didn't return a paramter negotioation.");
                }
                try {
                    return HdlcParameterNegotiation.decode(receiveFrame.informationField());
                } catch (FrameInvalidException e) {
                    throw new FatalJDlmsException(HDLC_CONNECTION_ESTABLISH_ERROR, SYSTEM,
                            "Received parameter negotiation, containes errors. Evaluate cause for details.", e);
                }

            default:
            case DISCONNECT_MODE:
                // remoteDisconnect(this); Close connection properly
                // dataExchangeConnection.close();
                close(settings);
                throw new FatalJDlmsException(HDLC_CONNECTION_ESTABLISH_ERROR, SYSTEM,
                        MessageFormat.format("Received a {0} frame, while connecting. Connection has been shot down.",
                                receiveFrame.frameType()));
            }
        }

        private void registerNewListener(HdlcAddressPair key, HdlcConnectionListener listener) {
            synchronized (this.listeners) {
                this.listeners.put(key, listener);
            }
        }

        private class ConnectionReader implements Runnable {

            @Override
            public void run() {
                try {
                    mainLoop();
                } catch (SocketException | CommPortException e) {
                    // ignore
                } catch (IOException e) {
                    notifyAllListners(e);
                } finally {
                    closeAll();
                }
            }

            private void closeAll() {
                try {
                    transportLayer.close();
                    listeners.clear();
                } catch (IOException e) {
                    // ignore
                }
            }

            private void notifyAllListners(IOException e) {
                for (HdlcConnectionListener listener : listeners.values()) {
                    listener.connectionInterrupted(e);
                }
            }

            private void mainLoop() throws IOException {
                while (!transportLayer.isClosed()) {
                    transportLayer.setTimeout(0);
                    List<HdlcFrame> decode = HdlcMessageDecoder.decode(transportLayer,
                            settings.messageFragmentTimeout());
                    for (HdlcFrame hdlcFrame : decode) {
                        HdlcAddressPair switchedPair = hdlcFrame.addressPair().switchedPair();
                        // System.out.println(hdlcFrame.frameType());
                        // System.out.printf("A: %s B:%s\n", connectionKey, switchedPair);
                        if (connectionKey != null && connectionKey.equals(switchedPair)) {
                            try {
                                incommingQueue.put(hdlcFrame);
                            } catch (InterruptedException e) {
                                // TODO: handle this properly??
                            }
                        }
                        else {
                            synchronized (listeners) {
                                HdlcConnectionListener listener = listeners.get(switchedPair);

                                if (listener == null) {
                                    throw new FatalJDlmsException(ExceptionId.HDLC_KEY_PAIR_ERR, Fault.SYSTEM,
                                            "Server replied with an unknown keypair.");
                                }

                                listener.dataReceived(hdlcFrame);
                            }
                        }

                    }
                }
            }

        }

    }

    public interface HdlcConnectionListener {
        void dataReceived(HdlcFrame frame);

        void connectionInterrupted(IOException e);
    }

}
