/*
 * Copyright 2012-15 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms.internal.sessionlayer.hdlc;

import static org.openmuc.jdlms.JDlmsException.ExceptionId.CONNECTION_ALREADY_CLOSED;
import static org.openmuc.jdlms.JDlmsException.Fault.SYSTEM;
import static org.openmuc.jdlms.internal.sessionlayer.hdlc.HdlcParameterNegotiation.MAX_WINDOW_SIZE;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.SocketTimeoutException;
import java.nio.ByteBuffer;
import java.text.MessageFormat;
import java.util.Arrays;

import org.openmuc.jdlms.FatalJDlmsException;
import org.openmuc.jdlms.JDlmsException.ExceptionId;
import org.openmuc.jdlms.JDlmsException.Fault;
import org.openmuc.jdlms.internal.ConfirmedMode;
import org.openmuc.jdlms.internal.connectionsettings.HdlcSettings;
import org.openmuc.jdlms.internal.sessionlayer.SessionLayer;
import org.openmuc.jdlms.internal.sessionlayer.SessionLayerListener;
import org.openmuc.jdlms.internal.sessionlayer.hdlc.HdlcDispatcher.HdlcConnection;
import org.openmuc.jdlms.internal.sessionlayer.hdlc.HdlcDispatcher.HdlcConnectionListener;

import gnu.io.CommPortTimeoutException;

/*
 * See IEC 62056-46 for further details.
 */
public class HdlcLayer implements SessionLayer {

    private static byte[] LLC_REQUEST = new byte[] { (byte) 0xE6, (byte) 0xE6, (byte) 0x00 };

    private SessionLayerListener connectionListener;

    private final HdlcSettings settings;

    private int sendSequence;
    private int receiveSequence;

    private HdlcMessageQueue sendQueue;

    private int sendWindowSize;
    private int sendInformationLength;

    private final ByteArrayOutputStream segmentBuffer;

    private boolean closed;

    private final HdlcConnection hdlcConnection;

    public HdlcLayer(HdlcSettings settings) {
        this.segmentBuffer = new ByteArrayOutputStream();
        this.settings = settings;

        this.sendSequence = 0;
        this.receiveSequence = 0;

        this.sendQueue = new HdlcMessageQueue(1);

        this.closed = true;

        this.hdlcConnection = HdlcDispatcher.instance().connect(settings, new HdlcConnectionListenerImpl());
    }

    @Override
    public synchronized void startListening(SessionLayerListener listener) throws IOException {
        connectionListener = listener;

        HdlcParameterNegotiation parameterNegotiation = this.hdlcConnection.open(this.settings);

        this.sendInformationLength = parameterNegotiation.receiveInformationLength() - LLC_REQUEST.length;
        this.sendWindowSize = parameterNegotiation.receiveWindowSize();

        HdlcMessageQueue oldQueue = this.sendQueue;
        if (oldQueue.capacity() < sendWindowSize) {
            this.sendQueue = new HdlcMessageQueue(sendWindowSize, oldQueue);
        }

        this.closed = false;
    }

    @Override
    public synchronized void send(byte[] tSdu, int off, int len) throws IOException {
        byte[] data = Arrays.copyOfRange(tSdu, off, off + len);
        check(len);

        if (len > sendInformationLength) {
            byte[] segment = new byte[sendInformationLength + LLC_REQUEST.length];
            segment = addLlcToFrame(segment);

            ByteBuffer dataWrapper = ByteBuffer.wrap(data);

            while (dataWrapper.remaining() > sendInformationLength) {
                dataWrapper.get(segment, LLC_REQUEST.length, segment.length - LLC_REQUEST.length);
                send(segment, true);
            }

            dataWrapper.get(segment, LLC_REQUEST.length, dataWrapper.remaining());
            send(segment, false);
        }
        else {
            byte[] frame = new byte[data.length + LLC_REQUEST.length];
            frame = addLlcToFrame(frame);

            System.arraycopy(data, 0, frame, LLC_REQUEST.length, data.length);
            send(frame, false);
        }
    }

    @Override
    public synchronized void close() throws IOException {
        if (this.closed) {
            throw new FatalJDlmsException(CONNECTION_ALREADY_CLOSED, Fault.USER, "Connection has been already closed.");
        }
        try {
            this.hdlcConnection.close(settings);
        } finally {
            closed = true;
            sendSequence = 0;
            receiveSequence = 0;
            sendQueue.clear();
        }
    }

    private void closeUnsafe() {
        try {
            close();
        } catch (IOException e) {
            // ignore
        }
    }

    private byte[] addLlcToFrame(byte[] frame) {
        System.arraycopy(LLC_REQUEST, 0, frame, 0, LLC_REQUEST.length);

        return frame;
    }

    private void check(int dataLength) throws IOException {
        if (sendQueue.size() >= MAX_WINDOW_SIZE) {
            throw new FatalJDlmsException(ExceptionId.HDLC_SEND_FRAME_SIZE_EXCEEDED, SYSTEM, "Send queue full");
        }
        int maxFrameLength = sendInformationLength * sendWindowSize;
        if (dataLength > maxFrameLength) {
            throw new FatalJDlmsException(ExceptionId.HDLC_SEND_FRAME_SIZE_EXCEEDED, SYSTEM, MessageFormat
                    .format("Message too large. {0} bytes allowed. Tried to send {1}", maxFrameLength, dataLength));
        }
    }

    private void send(byte[] data, boolean segmented) throws IOException {
        HdlcAddressPair addressPair = settings.addressPair();
        HdlcFrame frame;

        if (settings.confirmedMode() == ConfirmedMode.CONFIRMED) {
            frame = HdlcFrame.newInformationFrame(addressPair, nextSendSequenceNumber(), currentReceiveSeqNumber(),
                    data, segmented);
        }
        else {
            frame = HdlcFrame.newUnnumberedInformationFrame(addressPair, data, false);
        }

        sendAndBufferFrame(frame);
    }

    private void sendAcknowledge() throws IOException {
        HdlcFrame frame = HdlcFrame.newReceiveReadyFrame(settings.addressPair(), receiveSequence, true);

        try {
            this.hdlcConnection.send(frame.encode());
        } catch (FrameInvalidException e) {
            // ignore frame Invalid..
        } catch (CommPortTimeoutException | SocketTimeoutException e) {
            // ignore this
        }
    }

    private void acknowledgeSendFramesTil(int sendSeq) {
        sendSeq = sendSeq == 0 ? 8 : sendSeq;

        if (!sendQueue.containsSequenceNumber(sendSeq)) {
            return;
        }

        sendQueue.clearTil(sendSeq);
    }

    private int nextSendSequenceNumber() {
        int result = sendSequence;
        sendSequence = (++sendSequence) % 8;
        return result;
    }

    private void incrementReceiveSequenceNumber() {
        receiveSequence = (++receiveSequence) % 8;
    }

    private int currentReceiveSeqNumber() {
        return receiveSequence;
    }

    private void sendAndBufferFrame(HdlcFrame frame) throws IOException {
        try {
            byte[] dataToSend = frame.encode();

            sendQueue.offerMessage(dataToSend, frame.sendSequence());

            this.hdlcConnection.send(dataToSend);

        } catch (FrameInvalidException e) {
        }
    }

    private void bufferSegment(HdlcFrame segment) {
        try {
            segmentBuffer.write(segment.informationField());
        } catch (IOException e) {
            // TODO
            // LoggingHelper.logStackTrace(e, logger);
        }
    }

    private boolean hasSegmentBuffered() {
        return segmentBuffer.size() > 0;
    }

    private byte[] clearBufferedSegment() {
        byte[] segment = segmentBuffer.toByteArray();
        segmentBuffer.reset();
        return segment;
    }

    private class HdlcConnectionListenerImpl implements HdlcConnectionListener {

        @Override
        public void dataReceived(HdlcFrame frame) {
            incrementReceiveSequenceNumber();
            if (frame.segmented()) {
                bufferSegment(frame);
                try {
                    sendAcknowledge();
                } catch (IOException e) {
                    closeUnsafe();
                    connectionInterrupted(e);
                    return;
                }
            }
            else if (frame.frameType() == FrameType.INFORMATION) {
                acknowledgeSendFramesTil(frame.receiveSequence());
                byte[] dlms;
                byte[] wholeFrame;
                if (hasSegmentBuffered()) {
                    bufferSegment(frame);
                    wholeFrame = clearBufferedSegment();
                }
                else {
                    wholeFrame = frame.informationField();
                }
                dlms = new byte[wholeFrame.length - 3];
                System.arraycopy(wholeFrame, 3, dlms, 0, dlms.length);
                connectionListener.dataReceived(dlms);
            }
            else if (frame.frameType() == FrameType.RECEIVE_READY) {
                acknowledgeSendFramesTil(frame.receiveSequence());
                sendRemainingFrames();
            }
        }

        @Override
        public void connectionInterrupted(IOException e) {
            connectionListener.connectionInterrupted(e);
        }

    }

    private void sendRemainingFrames() {
        try {
            for (byte[] message : sendQueue) {
                this.hdlcConnection.send(message);
            }
        } catch (IOException e) {
            closeUnsafe();
            this.connectionListener.connectionInterrupted(e);
        }

    }

}
