/*
 * Copyright 2012-16 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms.app;

import static java.lang.System.exit;
import static org.openmuc.jdlms.app.ConsoleClient.Action.PRINT_HELP;
import static org.openmuc.jdlms.app.ConsoleClient.Action.QUIT;
import static org.openmuc.jdlms.app.ConsoleClient.Action.READ;
import static org.openmuc.jdlms.app.ConsoleClient.Action.SCAN;
import static org.openmuc.jdlms.app.ConsoleClient.Action.WRITE;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

import org.openmuc.jdlms.internal.cli.CliParseException;

public final class ConsoleClient {

    private static final String LINES = "------------------------------------------------------";

    private static ConsoleApp consoleApp;

    static {
        Runtime.getRuntime().addShutdownHook(new Thread() {
            @Override
            public void run() {
                if (consoleApp != null) {
                    consoleApp.close();
                }
            }
        });
    }

    public static void main(String[] args) {
        start(args);
    }

    public static void start(String[] args) {

        ConsoleClientCliParser clientLineParser = new ConsoleClientCliParser();
        try {
            clientLineParser.parse(args);
            consoleApp = clientLineParser.connectAndCreateConsoleApp();
        } catch (CliParseException e) {
            System.err.println("Error parsing command line parameters: " + e.getMessage());
            clientLineParser.printUsage();
            System.exit(1);
        } catch (IOException e) {
            System.err.println("Error connecting to meter: " + e.getMessage());
            e.printStackTrace();
            System.exit(1);
        }

        processActions();
    }

    private static void processActions() {
        printHelp();

        BufferedReader inputReader = new BufferedReader(new InputStreamReader(System.in));
        try {

            String actionStr;
            while (true) {
                System.out.println("\n** Enter actionkey: ");

                try {
                    actionStr = inputReader.readLine();
                } catch (IOException e) {
                    System.err.printf("%s. Application is being shut down.%n", e.getMessage());
                    exit(2);
                    return;
                }

                Action action;
                try {
                    action = Action.actionFor(actionStr);
                } catch (IllegalArgumentException e) {
                    System.err.println("Illegal actionkey.\n");

                    printHelp();
                    continue;
                }

                switch (action) {
                case READ:
                    consoleApp.processRead();
                    break;

                case WRITE:
                    consoleApp.processWrite();
                    break;

                case SCAN:
                    System.out.println("** Scan started...");
                    consoleApp.processScan();
                    break;

                case QUIT:
                    System.out.println("** Closing connection.");
                    return;

                case PRINT_HELP:
                default:
                    printHelp();
                    break;
                }

            }

        } catch (IOException e) {
            System.err.println("Connection closed for the following reason: " + e.getMessage());
            e.printStackTrace(); // TODO: remove this.
            return;
        } catch (Exception e) {
            e.printStackTrace();
            return;
        } finally {
            consoleApp.close();
        }
    }

    private static void printHelp() {
        String message = " %s - to %s\n";
        System.out.flush();
        System.out.println();
        System.out.println(LINES);
        System.out.printf(message, QUIT.key(), "quit the connection");
        System.out.println(LINES);

        System.out.printf(message, READ.key(), "read");
        System.out.printf(message, WRITE.key(), "write");
        System.out.printf(message, SCAN.key(), "scan");

        System.out.println(LINES);
        System.out.printf(" %s - for help\n", PRINT_HELP.key());
        System.out.println(LINES);

    }

    enum Action {
        READ("1"),
        WRITE("2"),
        SCAN("3"),

        PRINT_HELP("?"),
        QUIT("q");

        private final String key;

        private static final Action[] ACTIONS = Action.values();

        private Action(String key) {
            this.key = key;
        }

        public String key() {
            return this.key;
        }

        public static Action actionFor(String key) throws IllegalArgumentException {

            for (Action action : ACTIONS) {
                if (action.key.equals(key)) {
                    return action;
                }
            }
            throw new IllegalArgumentException("Wrong key!");
        }
    }

}
